import dataclasses
import logging
import re
from collections import OrderedDict
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

logger = logging.getLogger(__name__)


class RowWrapper:
    """
    Build lightweight row tuples for DB API and csv.DictReader rows.

    Inspired by Greg Stein's lovely
    `dtuple module <https://code.activestate.com/recipes/81252-using-dtuple-for-flexible-query-result-access>`_,
    which I can't find online any longer, isn't on pypi, and doesn't support Python 3 without some fixes.

    Initializer takes a sequence of column descriptions, either names, or tuples of names and other metadata (which
    will be ignored). For instance, it's happy to take a DB API cursor description, or a csv.DictReader's fieldnames
    property. Provides a wrap(row) method for wrapping rows.

    Some characters which are illegal in identifiers will be replaced when building the row tuples - currently "-" and
    " " characters will be replaced with "_"s.

    >>> cursor = conn.cursor()
    >>> cursor.execute("SELECT kind, rating FROM sausages ORDER BY rating DESC;")
    >>> wrapper = RowWrapper(cursor.description)
    >>> rows = [wrapper.wrap(row) for row in cursor.fetchall()]

    >>> reader = csv.DictReader(csv_file)
    >>> wrapper = RowWrapper(reader.fieldnames)
    >>> rows = [wrapper.wrap(row) for row in reader]
    """

    def __init__(
        self,
        description: Sequence[str | tuple[str]],
        force_lower_case_ids: bool = False,
        row_tuple_class_name="Row",
    ) -> None:
        column_names = description if isinstance(description[0], str) else [col[0] for col in description]
        self.ids_and_column_names = self._ids_and_column_names(column_names, force_lower_case=force_lower_case_ids)
        self.dataclass = dataclasses.make_dataclass(row_tuple_class_name, self.ids_and_column_names.keys())

    @staticmethod
    def _ids_and_column_names(names, force_lower_case=False):
        """Ensure all column names are unique identifiers."""
        fixed = OrderedDict()
        for name in names:
            identifier = RowWrapper._make_identifier(name)
            if force_lower_case:
                identifier = identifier.lower()
            while identifier in fixed:
                identifier = RowWrapper._increment_numeric_suffix(identifier)
            fixed[identifier] = name
        return fixed

    @staticmethod
    def _make_identifier(string):
        """Attempt to convert string into a valid identifier by replacing invalid characters with "_"s,
        and prefixing with "a_" if necessary."""
        string = re.sub(r"[ \-+/\\*%&$£#@.,;:'" "?<>]", "_", string)
        if re.match(r"^\d", string):
            string = f"a_{string}"
        return string

    @staticmethod
    def _increment_numeric_suffix(s):
        """Increment (or add) numeric suffix to identifier."""
        if re.match(r".*\d+$", s):
            return re.sub(r"\d+$", lambda n: str(int(n.group(0)) + 1), s)
        return s + "_2"

    def wrap(self, row: Mapping[str, Any] | Sequence[Any]):
        """Return row tuple for row."""
        return (
            self.dataclass(**{ident: row[column_name] for ident, column_name in self.ids_and_column_names.items()})
            if isinstance(row, Mapping)
            else self.dataclass(**dict(zip(self.ids_and_column_names.keys(), row, strict=False)))
        )

    def wrap_all(self, rows: Iterable[Mapping[str, Any] | Sequence[Any]]):
        """Return row tuple for each row in rows."""
        return (self.wrap(r) for r in rows)

    def __call__(self, row):
        return self.wrap(row)



import logging
import sqlite3
from contextlib import closing
from io import StringIO

import pytest

logger = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def db():
    with closing(sqlite3.connect(":memory:")) as conn:
        cursor = conn.cursor()

        cursor.execute("CREATE TABLE sausages (kind VARCHAR NOT NULL PRIMARY KEY, rating INT NOT NULL);")
        cursor.execute("INSERT INTO sausages VALUES (?, ?);", ("cumberland", 10))
        cursor.execute("INSERT INTO sausages VALUES (?, ?);", ("vegetarian", 0))
        cursor.execute("INSERT INTO sausages VALUES (?, ?);", ("lincolnshire", 9))
        conn.commit()

        yield conn


@pytest.fixture(scope="function")
def csv_file():
    data = "kind,rating\ncumberland,10\nlincolnshire,9\nvegetarian,0\n"
    return StringIO(data)



import logging

from hamcrest import assert_that, has_properties

from brunns.row.rowwrapper import RowWrapper

logger = logging.getLogger(__name__)


def test_identifiers_fixed_for_mapping_row():
    # Given
    wrapper = RowWrapper(["column-name", "Another One", "3rd Column"])

    # When
    row = wrapper({"column-name": "value", "Another One": "another-value", "3rd Column": "3rd value"})

    # Then
    assert_that(
        row,
        has_properties(column_name="value", Another_One="another-value", a_3rd_Column="3rd value"),
    )


def test_identifiers_fixed_for_positional_row():
    # Given
    wrapper = RowWrapper(["column-name", "Another One", "3rd Column"])

    # When
    row = wrapper(["value", "another", "yet another"])

    # Then
    assert_that(row, has_properties(column_name="value", Another_One="another", a_3rd_Column="yet another"))


def test_column_identifiers_deduplication_for_mapping_row():
    # Given
    wrapper = RowWrapper(["column-name", "column$name"])

    # When
    row = wrapper({"column-name": "value", "column$name": "another-value"})

    # Then
    assert_that(row, has_properties(column_name="value", column_name_2="another-value"))


def test_column_identifiers_deduplication_for_positional_row():
    # Given
    wrapper = RowWrapper(["column-name", "column-name", "column$name"])

    # When
    row = wrapper(["value", "another", "yet another"])

    # Then
    assert_that(
        row,
        has_properties(column_name="value", column_name_2="another", column_name_3="yet another"),
    )


def test_lower_cased_identifiers():
    # Given
    wrapper = RowWrapper(["column-name", "Another One", "3rd Column"], force_lower_case_ids=True)

    # When
    row = wrapper({"column-name": "value", "Another One": "another-value", "3rd Column": "3rd value"})

    # Then
    assert_that(
        row,
        has_properties(column_name="value", another_one="another-value", a_3rd_column="3rd value"),
    )



import csv
import logging

from hamcrest import assert_that, contains_exactly, has_properties

from brunns.row.rowwrapper import RowWrapper

logger = logging.getLogger(__name__)


def test_dbapi_row_wrapping(db):
    # Given
    cursor = db.cursor()
    cursor.execute("SELECT kind, rating FROM sausages ORDER BY rating DESC;")

    # When
    wrapper = RowWrapper(cursor.description)
    rows = [wrapper.wrap(row) for row in cursor.fetchall()]

    # Then
    assert_that(
        rows,
        contains_exactly(
            has_properties(kind="cumberland", rating=10),
            has_properties(kind="lincolnshire", rating=9),
            has_properties(kind="vegetarian", rating=0),
        ),
    )


def test_wrap_all(db):
    # Given
    cursor = db.cursor()
    cursor.execute("SELECT kind, rating FROM sausages ORDER BY rating DESC;")

    # When
    wrapper = RowWrapper(cursor.description)
    rows = wrapper.wrap_all(cursor.fetchall())

    # Then
    assert_that(
        rows,
        contains_exactly(
            has_properties(kind="cumberland", rating=10),
            has_properties(kind="lincolnshire", rating=9),
            has_properties(kind="vegetarian", rating=0),
        ),
    )


def test_csv_wrapping(csv_file):
    # Given
    reader = csv.DictReader(csv_file)

    # When
    wrapper = RowWrapper(reader.fieldnames)
    rows = [wrapper.wrap(row) for row in reader]

    # Then
    assert_that(
        rows,
        contains_exactly(
            has_properties(kind="cumberland", rating="10"),
            has_properties(kind="lincolnshire", rating="9"),
            has_properties(kind="vegetarian", rating="0"),
        ),
    )
